#include <time/Clock.hpp>

#if defined(OS_LINUX)
# include <sys/time.h>
#elif defined(OS_WINDOWS)
# include <windows.h>
# include <mmsystem.h>
# ifdef COMPILER_MSVC
#  pragma comment (lib, "winmm")
# endif
#endif

#include <ctime>
using namespace std;
using namespace cat;

Clock::Clock()
{
#if defined(OS_WINDOWS)
	// Set minimum timer resolution as low as possible
	for (period = 1; period < 20 && (timeBeginPeriod(period) != TIMERR_NOERROR); ++period);
#endif
}

Clock::~Clock()
{
#if defined(OS_WINDOWS)
	if (period < 20) timeEndPeriod(period);
#endif
}


u32 Clock::sec()
{
#if defined(OS_LINUX)

	struct timeval cateq_v;
	struct timezone cateq_z;

	gettimeofday(&cateq_v, &cateq_z);

	return static_cast<u32>(cateq_v.tv_sec);

#elif defined(OS_WINDOWS)

	return timeGetTime() / 1000;

#endif
}

u32 Clock::hsec()
{
#if defined(OS_LINUX)

	struct timeval cateq_v;
	struct timezone cateq_z;

	gettimeofday(&cateq_v, &cateq_z);

	return static_cast<u32>(100.0 * static_cast<double>(cateq_v.tv_sec) + static_cast<double>(cateq_v.tv_usec) / 10000.0);

#elif defined(OS_WINDOWS)

	return timeGetTime() / 10;

#endif
}

u32 Clock::msec()
{
#if defined(OS_LINUX)

	struct timeval cateq_v;
	struct timezone cateq_z;

	gettimeofday(&cateq_v, &cateq_z);

	return static_cast<u32>(1000.0 * static_cast<double>(cateq_v.tv_sec) + static_cast<double>(cateq_v.tv_usec) / 1000.0);

#elif defined(OS_WINDOWS)

	return timeGetTime();

#endif
}

double Clock::usec()
{
#if defined(OS_LINUX)

	struct timeval cateq_v;
	struct timezone cateq_z;

	gettimeofday(&cateq_v, &cateq_z);

	return 1000000.0 * static_cast<double>(cateq_v.tv_sec) + static_cast<double>(cateq_v.tv_usec);

#elif defined(OS_WINDOWS)

	/* In Windows, this value can leap forward randomly:
	 * http://support.microsoft.com/default.aspx?scid=KB;EN-US;Q274323
	 * So it may be best to do most of the timing in milliseconds.
	 */

	LARGE_INTEGER tim, freq; // 64-bit! ieee
	double usec;

	QueryPerformanceCounter(&tim);
	QueryPerformanceFrequency(&freq);
	usec = (static_cast<double>(tim.QuadPart) * 1000000.0) / static_cast<double>(freq.QuadPart);

	return usec;

#endif
}


void Clock::sleep(u32 milliseconds)
{
#if defined(OS_LINUX)
	usleep(milliseconds * 1000);
#elif defined(OS_WINDOWS)
	Sleep(milliseconds);
#endif
}


std::string Clock::format(const char *format_string)
{
	char ts[256];

	struct tm localTime;

	__time64_t long_time;
	_time64(&long_time);
	_localtime64_s(&localTime, &long_time);

	strftime(ts, sizeof(ts), format_string, &localTime);

	return ts;
}
