/*
	Unit test for the Elliptic Curve Cryptography code

	Performs 10,000 simulated key exchanges, terminating on any error

	Demonstrates how to use the ECC code
*/

#include <crypto/publickey/ECC.hpp>
#include <math/BigInt.hpp>
#include <time/Clock.hpp>
#include <rand/Random.hpp>
#include <iostream>
using namespace std;
using namespace cat;

// Generate candidate values for c for the ECC.cpp code
void GenerateCandidatePrimes();

// Generate table used for w-MOF in the ECC.cpp code
void GenerateMOFTable(int window_bits);

// Stress testing the modular square root code
void TestSquareRoot();

// Compare Barrett, Montgomery, and Special modulus functions for speed
void TimeModulusCode();

// ECC protocol test
void ECCTest();

int main()
{
	cout << "Candidate primes for ECC:" << endl;
	GenerateCandidatePrimes();

	cout << endl << "w-MOF table generation:" << endl;
	GenerateMOFTable(4);

	cout << endl << "Square Root testing:" << endl;
	TestSquareRoot();

	cout << endl << "Modulus code timing:" << endl;
	TimeModulusCode();

	cout << endl << "ECC testing:" << endl;
	ECCTest();

	return 0;
}

void GenerateCandidatePrimes()
{
	u32 k[EDWARD_LIMBS];

	for (int ii = 1; ii < 1000; ii += 4)
	{
		Set32(k, EDWARD_LIMBS, 0);
		Subtract32(k, EDWARD_LIMBS, ii);

		if (RabinMillerPrimeTest(k, EDWARD_LIMBS, 40))
		{
			cout << "Candidate prime c = " << ii << endl;
		}
	}
}

void GenerateMOFTable(int window_bits)
{
	cout << "When we see each combinations of w+1 bits, what operations should be performed?" << endl;
	cout << "It will be a number of doubles, then an addition by an odd number, then some more doublings." << endl;

	int top = 1 << (window_bits + 1);
	int cnt = 0;

	for (int ii = 0; ii < top; ++ii)
	{
		int r = 0;

		for (int jj = window_bits; jj >= 1; --jj)
		{
			int top = ii & (1 << (jj-1));
			int bottom = ii & (1 << jj);

			if (top)
			{
				if (bottom)
				{
					// 0
				}
				else
				{
					// 1
					r += 1 << (jj - 1);
				}
			}
			else
			{
				if (bottom)
				{
					// -1
					r -= 1 << (jj - 1);
				}
				else
				{
					// 0
				}
			}
		}

		int squares_before = window_bits;
		int squares_after = 0;
		if (r)
		{
			while (!(r&1))
			{
				--squares_before;
				++squares_after;
				r >>= 1;
			}
		}

		cout << ii << "(";

		for (int jj = 1 << window_bits; jj; jj >>= 1)
		{
			if (ii & jj)	cout << '1';
			else			cout << '0';
		}

		cout << ") -> " << squares_before << "D + (" << r << ") + " << squares_after << "D" << endl;
/*
		if (ii & 1)
		{
			if (r >= 0)
			{
				cout << "{" << ((r + 1) >> 1) - 1 << "," << squares_after << "},";
				if (++cnt == 8)
				{
					cnt = 0;
					cout << endl;
				}
			}
		}*/
	}
}

void TestSquareRoot()
{
	u32 k[EDWARD_LIMBS];
	u32 l[EDWARD_LIMBS];
	const u32 c = 189;

	//for (int ii = 0; ii < 100000; ++ii)
	{
		Random::ref()->generate(k, sizeof(k));
		SpecialModulus(k, EDWARD_LIMBS, c, EDWARD_LIMBS, k);

		u32 p[EDWARD_LIMBS*2];
		Square(EDWARD_LIMBS, p, k);
		SpecialModulus(p, EDWARD_LIMBS*2, c, EDWARD_LIMBS, l);

		SpecialSquareRoot(EDWARD_LIMBS, l, c, l);

		if (Equal(EDWARD_LIMBS, k, l))
		{
			cout << "Okay for " << ToStr(k, EDWARD_LIMBS, 16) << endl;
		}
		else
		{
			Negate(EDWARD_LIMBS, l, l);
			Subtract32(l, EDWARD_LIMBS, c);

			if (Equal(EDWARD_LIMBS, k, l))
			{
				cout << "Okay for " << ToStr(k, EDWARD_LIMBS, 16) << endl;
			}
			else
			{
				cout << "Fail for " << ToStr(k, EDWARD_LIMBS, 16) << endl;
				return;
			}
		}
	}
}

void TimeModulusCode()
{
#define LIMBS 8
	u32 a[LIMBS*2], b[LIMBS], modulus[LIMBS];
	u32 p_x[LIMBS], p_y[LIMBS];
	u32 three[LIMBS];

	Set32(three, LIMBS, 3);
	Set32(b, LIMBS, 11);
	Set32(p_x, LIMBS, 8);
	Set32(p_y, LIMBS, 15);

	//INFO("Test") << "Generating a pseudo-prime modulus...";
	//GenerateStrongPseudoPrime(modulus, LIMBS);
	Set32(modulus, LIMBS, 0);
	const u32 C = 189;
	Subtract32(modulus, LIMBS, C);
	cout << "Modulus: " << ToStr(modulus, LIMBS, 16) << endl;

	u32 v[LIMBS*2], v_inverse[LIMBS];
	u32 p[LIMBS*2+1], presidue[LIMBS*2+1], q[LIMBS*2], r[LIMBS];

	Set(v, LIMBS*2, modulus, LIMBS);
	BarrettModulusPrecomp(LIMBS, modulus, v_inverse);

	cout << "v_inverse = " << ToStr(v_inverse, LIMBS, 16) << endl;

	double ac1 = 0, ac2 = 0, ac3 = 0;
	for (int ii = 0; ii < 100000; ++ii)
	{
		Random::ref()->generate(a, LIMBS*4);
		Random::ref()->generate(b, LIMBS*4);
		Multiply(LIMBS, p, a, b);

		MonInputResidue(p, LIMBS*2, v, LIMBS, presidue);
		u32 mod_inv = MonReducePrecomp(v[0]);

		double clk1 = Clock::usec();
		BarrettModulus(LIMBS, p, v, v_inverse, q);
		double clk2 = Clock::usec();
		MonFinish(LIMBS, presidue, v, mod_inv);
		double clk3 = Clock::usec();
		SpecialModulus(p, LIMBS*2, C, LIMBS, r);
		double clk4 = Clock::usec();

		double t3 = clk4 - clk3;
		double t2 = clk3 - clk2;
		double t1 = clk2 - clk1;

		ac1 += t1;
		ac2 += t2;
		ac3 += t3;

		for (int ii = 0; ii < LIMBS; ++ii)
		{
			if (presidue[ii] != q[ii])
			{
				cout << "FAIL: Montgomery doesn't match Barrett" << endl;
				break;
			}
			if (presidue[ii] != r[ii])
			{
				cout << "FAIL: Montgomery doesn't match Special" << endl;
				cout << "Expected = " << ToStr(q, LIMBS, 16) << endl;
				cout << "Got      = " << ToStr(r, LIMBS, 16) << endl;
				Subtract(q, LIMBS, r, LIMBS);
				cout << "Difference = " << ToStr(q, LIMBS, 16) << endl;
				u32 crem = Divide32(LIMBS, q, C);
				cout << "Delta Divides C = " << ToStr(q, LIMBS, 16) << " times" << endl;
				cout << "Delta Modulus C = " << crem << endl;
				break;
			}
		}
	}

	cout << "MonFinish()/SpecialModulus(): " << ac2 / ac3 << endl;
	cout << "BarrettModulus()/MonFinish(): " << ac1 / ac2 << endl;
}

void ECCTest()
{
	//for (int ii = 0; ii < 10000; ++ii)
	{
		TwistedEdwardServer server;
		TwistedEdwardClient client;

		EdPrivateKey server_private_key;
		EdServerPublicKey server_public_key;
		EdClientPublicKey client_public_key;
		EdSharedSecret client_shared_secret;
		EdSharedSecret server_shared_secret;

		double t1, t2;

		cout << "OFFLINE PRECOMPUTATION: Generating server private and public keys" << endl;
		t1 = Clock::usec();
		server.GenerateOfflineStuff(&server_private_key, &server_public_key);
		t2 = Clock::usec();
		cout << "-- Operation completed in " << (t2 - t1) << " usec" << endl;

		cout << "ON SERVER START-UP: Loading server private key" << endl;
		t1 = Clock::usec();
		server.SetPrivateKey(&server_private_key);
		t2 = Clock::usec();
		cout << "-- Operation completed in " << (t2 - t1) << " usec" << endl;

		cout << "ON CLIENT START-UP: Loading server public key" << endl;
		t1 = Clock::usec();
		client.SetServerPublicKey(&server_public_key);
		t2 = Clock::usec();
		cout << "-- Operation completed in " << (t2 - t1) << " usec" << endl;

		cout << "CLIENT: ONLINE NOT TIME-CRITICAL: Computing shared secret" << endl;
		t1 = Clock::usec();
		client.ComputeSharedSecret(&client_public_key, &client_shared_secret);
		t2 = Clock::usec();
		cout << "-- Operation completed in " << (t2 - t1) << " usec" << endl;

		cout << "SERVER: ONLINE AND TIME-CRITICAL: Computing shared secret" << endl;
		t1 = Clock::usec();
		if (!server.ComputeSharedSecret(&client_public_key, &server_shared_secret))
		{
			cout << "FAILURE: Server rejected public key" << endl;
			return;
		}
		else
		{
			t2 = Clock::usec();
			cout << "SUCCESS: Server accepted public key" << endl;
		}
		cout << "-- Operation completed in " << (t2 - t1) << " usec" << endl;

		if (memcmp(server_shared_secret.Bytes, client_shared_secret.Bytes, sizeof(server_shared_secret.Bytes)))
		{
			cout << "FAILURE: Server shared secret did not match client shared secret" << endl;
			return;
		}
		else
		{
			cout << "SUCCESS: Server shared secret matched client shared secret" << endl;
		}
	}
}
